@@ -30,17 +30,15 @@
 #include <stdlib.h>
 #include <signal.h>
 #include <string.h>             /* for strerror() */
-#include <termios.h>
 #include <unistd.h>
 
 #include <glib.h>
 
-static GSList *machines;
-static GHashTable *machine_names;
-static GHashTable *machine_fds;
+#include "ttywatch.h"
 
 /* configuration */
 static char *logpath;
+static char *modname;
 static char *configfile;
 static char *pidfile;
 static int intarg;
@@ -60,21 +58,17 @@ static struct poptOption optionsTable[] = {
       "signaling rate for current maching in bps", "<BPS>" },
     { "logpath", 'l', POPT_ARG_STRING, &logpath, 0,
       "directory path for logfiles", "path" },
+    { "module", 'm', POPT_ARG_STRING, &modname, 'm',
+      "path to module for logging this port", "path" },
     { "daemon", 'd', POPT_ARG_NONE, &do_daemon, 0,
       "directory path for logfiles", "path" },
       POPT_AUTOHELP
     { NULL, 0, 0, NULL, 0 }
 };
 
-typedef struct machine {
-	char *name;
-	char *devpath;
-	int dev_fd;
-	int log_fd;
-	speed_t speed;
-	struct termios ots;
-	struct termios ts;
-} machine;
+static GSList *machines;
+static GHashTable *machine_names;
+static GHashTable *machine_fds;
 
 void warn(char *warning, machine *m) {
 	char * message;
@@ -83,7 +77,7 @@ void warn(char *warning, machine *m) {
 	if (!m) abort();
 
 	message = alloca(strlen(warning) + strlen(m->name) + 15);
-	size = sprintf(message, "ttywatch[%s]: %s\n", m->name, message);
+	size = sprintf(message, "ttywatch[%s]: %s\n", m->name, warning);
 	write(STDERR_FILENO, message, size);
 	if (m->log_fd >= 0) write(m->log_fd, message, size);
 }
@@ -161,6 +155,9 @@ void delete_pidfile(void) {
 void open_one_logfile(machine *m) {
 	char *logname;
 
+	/* no logfile if a module is being used */
+	if (m->process_line) return;
+
 	logname = alloca(strlen(logpath)+strlen(m->name)+6);
 	sprintf(logname, "%s/%s.log", logpath, m->name);
 	m->log_fd = open(logname, O_RDWR|O_CREAT|O_APPEND, 0666);
@@ -220,7 +217,8 @@ void cleanup_one_termios(gpointer data, gpointer ignore) {
 
 	tcsetattr(m->dev_fd, TCSANOW, &m->ots);
 	close(m->dev_fd);
-	close(m->log_fd);
+	/* no logfile if a module is being used */
+	if (m->log_fd) close(m->log_fd);
 }
 
 /* restore all original terminal settings on exit */
@@ -233,7 +231,8 @@ void cleanup_termios_and_exit(int signal) {
 void reopen_one_logfile(gpointer data, gpointer ignore) {
 	machine *m = data;
 
-	close(m->log_fd);
+	/* no logfile if a module is being used */
+	if (m->log_fd) close(m->log_fd);
 	open_one_logfile(m);
 }
 
@@ -277,6 +276,12 @@ machine *parse_argument (char c, machine *m) {
 		if (!m) die ("--name or -n must come before --bps or -b\n");
 		m->speed = symbolic_speed(intarg);
 		break;
+	case 'm':
+		if (!m) die ("--name or -n must come before --module or -m\n");
+		if (m->modhandle) dlclose(m->modhandle);
+		m->modhandle = dlopen(chararg, RTLD_NOW);
+		m->process_line = dlsym(m->modhandle, "process_line");
+		break;
 	case 'c':
 		/* read the whole config file in at once */
 		conf_fd = open(configfile, O_RDONLY);
@@ -352,8 +357,10 @@ int main(int argc, char **argv) {
 	/* set everything up -- daemon() after setup to report problems */
 	setup_pidfile();
 	setup_termios();
-	if (do_daemon)
+	if (do_daemon) {
 		daemon(0, 0);
+		setup_pidfile(); /* daemon() has changed pid ... */
+	}
 
 	/* set the signal handler to restore the old
 	 * termios handler */
@@ -380,8 +387,14 @@ int main(int argc, char **argv) {
 				m = g_hash_table_lookup(machine_fds, &ufds[n].fd);
 				if (!m) continue;
 
-				i = read(m->dev_fd, buf, BUFSIZE);
-				if (i >= 1) write(m->log_fd, buf, i);
+                                i = read(m->dev_fd, buf, BUFSIZE-1);
+				if (i >= 1) {
+					buf[i+1] = '\0';
+					if (m->process_line)
+						m->process_line(m, buf);
+					else
+						write(m->log_fd, buf, i);
+				}
 			}
 		}
 	} while (1); /* exits through signal handler */
